# The description below is primarily based on Abramenko & Brown, Example 7.136, p. 434ff
# (which only deals with odd characteristic, though), missing details were filled
# in with help of Bruhat-Tits "Groupes réductifs sur un corps local", Section 10,
# and Abramenko's "Twin buildings and Applications to S-arithmetic groups"


# Apply field automorphism sigma to every entry in mat
PowMatElts := function (mat, sigma)
    return List(mat, row->List(row, x -> x^sigma));
end;;

# K: field
# n: Witt index >= 1
# def: defect = dimension of V0, the anisotropic kernel (only 0, 1, 2 are allowed)
# sigma: 1 or q, if we are working over GF(q^2)
# eps: +1 or -1
#
# For our application, the following parameters are relevant:
# Sp6: UnipotentRootGroups(GF(q),   3, 0, 1, -1);
# O7:  UnipotentRootGroups(GF(q),   3, 1, 1,  1);
# O8:  UnipotentRootGroups(GF(q),   3, 2, 1,  1);
# U6:  UnipotentRootGroups(GF(q^2), 3, 0, q,  1);
# U7:  UnipotentRootGroups(GF(q^2), 3, 1, q,  1);
UnipotentRootGroups := function(K, n, def, sigma, eps)
    local K_hi, K_lo, m, B, Q, i, j, x, norms, RootGroup, fund_rgs, pos_rgs, U, sub;

    Assert(0, n >= 1);
    Assert(0, def in [0..2]);
    Assert(0, sigma = 1 or sigma^2 = Size(K));
    Assert(0, eps = 1 or eps = -1);

    # compute the dimension of the matrices
    m := 2*n + def;

    # the (sigma,eps)-quadratic form
    Q := NullMat(m, m, K);
    for i in [1..n] do
        Q[n+i+def][n-i+1] := One(K);
    od;
    if def = 0 then
        # Nothing to do
    elif def = 1 then
        #Q[n+1][n+1] := First(K, x -> x^sigma <> -eps*x );
        if Characteristic(K) = 2 and sigma = 1 then
            Q[n+1][n+1] := One(K);
        else
            Q[n+1][n+1] := First(K, x -> IsOne(eps*x + x^sigma) );
        fi;
    elif def = 2 then
        # Only occurs for O8, so to simplify things, we only support that for now
        Assert(0, sigma = 1 and eps = 1);
        Q[n+1][n+1] := One(K);
        Q[n+1][n+2] := One(K);
        if Characteristic(K) = 3 then
            Q[n+2][n+2] := 2 * One(K);
        else
            Q[n+2][n+2] := One(K);
        fi;
    fi;

    # the corresponding (sigma,eps)-Hermitian form
    B := eps * TransposedMat(Q) + PowMatElts(Q, sigma);

    # Verify B(x,y) = B(y,x)^sigma * eps
    Assert(0, B = PowMatElts(TransposedMat(B), sigma) * eps);

    # Verify that B is non-degenerate (unless char = 2 and def > 0)
    if Characteristic(K) = 2 and def > 0 then
        Assert(0, Length(NullspaceMat(B)) <= 1);
    else
        Assert(0, IsZero(NullspaceMat(B)));
    fi;


    # Construct the root group U_{ij} (where U_i = U_{ii}).
    RootGroup := function (i, j)
        local mi, mj, pi, pj, epsi, epsj, gens, l, v, f, k, mat, lambda, Qv, vcoeff, base;

        Assert(0, AbsInt(i) in [1..m]);
        Assert(0, AbsInt(j) in [1..m]);
        if i > 0 then
            pi := n+i+def; mi := n-i+1; epsi := 1;
        else
            pi := n+i+1; mi := n-i+def; epsi := eps;
        fi;

        if j > 0 then
            pj := n+j+def; mj := n-j+1; epsj := 1;
        else
            pj := n+j+1; mj := n-j+def; epsj := eps;
        fi;

        gens := [];
        if AbsInt(i) <> AbsInt(j) then
            # Easy case: |i| and |j| differ
            for lambda in Basis(K) do
                mat := IdentityMat(m, K);
                mat[pi][pj] := epsi * lambda;
                mat[mj][mi] := -epsj * lambda^sigma;
                Add(gens, mat);
            od;
        else
            # First add generators for v = f = 0.
            # For orthogonal groups in char 2, the following gens are *wrong*
            if not (Characteristic(K) = 2 and sigma = 1 and eps = 1) then
                base := Filtered(K, lambda->not IsZero(lambda)
                            and IsZero(lambda + eps*lambda^sigma));
                for lambda in base do
                    mat := IdentityMat(m, K);
                    mat[pi][mi] := lambda;
                    Assert(0, B=TransposedMat(mat)*B*PowMatElts(mat,sigma));
                    Add(gens, mat);
                od;
            fi;

            # Now add gens for v <> 0 (if any)
            for l in [1..def] do
                for vcoeff in Basis(K) do
                    # We have  f(x) := -epsi * <x,v> = -epsi * (x^T * B * v^sigma)
                    v := NullMat(1, m, K)[1];

                    v[n+l] := vcoeff;
                    # FIXME: unify char 2?
                    if Characteristic(K) = 2 then
                        f := -epsi*B*List(v,c->c^sigma);    # FIXME: char 2?
                        Qv := v*Q*List(v,c->c^sigma);
                        base := [- epsi * Qv ];
                    else
                        f := -epsi*B*List(v,c->c^sigma);
                        Qv := v*B*List(v,c->c^sigma);
                        base := [ First(K, lambda->lambda + eps*lambda^sigma = - epsi * Qv) ];
                        #base := Filtered(K, lambda->lambda + eps*lambda^sigma = - epsi * Qv);
                    fi;
                    Assert(0, 0<>Size(base));
                    for lambda in base do
                        mat := IdentityMat(m, K);
                        mat[pi][mi] := lambda;
                        for k in [1..def] do
                            mat[pi][n+k] := f[n+k];
                            mat[n+k][mi] := v[n+k];
                        od;
                        Assert(0, B=TransposedMat(mat)*B*PowMatElts(mat,sigma));
                        Add(gens, mat);
                    od;
                od;
            od;
        fi;

        return Group(gens);
    end;; # RootGroup


    # Get all "positive" root groups.
    pos_rgs := [];
    for i in [1..n] do
        for j in [1..i] do
            Add(pos_rgs, RootGroup(i, j));
            if i<>j then
                Add(pos_rgs, RootGroup(i, -j));
            fi;
        od;
    od;

    # Choose a subset of the root groups corresponding to fundamental system of roots.
    fund_rgs := [];
    Add(fund_rgs, RootGroup(1, 1));
    for i in [1..n-1] do
        Add(fund_rgs, RootGroup(i+1, i));
    od;

    U := Group(Union(List(fund_rgs, GeneratorsOfGroup)));

    # Verify that U really preserves the forms
    Assert(0, ForAll(GeneratorsOfGroup(U), g-> B=TransposedMat(g)*B*PowMatElts(g,sigma)));
#    Assert(0, ForAll(GeneratorsOfGroup(U), g-> Q=TransposedMat(g)*Q*PowMatElts(g,sigma)));

    # Compute size of the group, conveniently deduced from the set of all positive
    # root groups
#    SetSize(U, Product(List(pos_rgs, Size)));

    return [U, fund_rgs, pos_rgs, B, Q];

end;; # UnipotentGroup
